/*
 * copyright (c)2009 http://wamsoft.jp
 * zlib license
 */
#include "sqthread.h"
#include <sqstdio.h>

namespace sqobject {

extern const SQChar *getString(HSQUIRRELVM v, int idx);
extern int getParamCount(HSQUIRRELVM v);

bool
Thread::isWait()
{
	return _waitList.size() > 0 || _waitTimeout >= 0;
}

/**
 * @return YXbhƌ݊Ǘ̃XbhvĂ true
 */
bool
Thread::isSameThread(HSQUIRRELVM v)
{
	return _thread.isSameThread(v);
}

/**
 * IuWFNgɑ΂҂
 * @param target ҂Ώ
 * @return YIuWFNg҂Ăꍇ true
 */
bool
Thread::notifyObject(Object *target)
{
	bool find = false;
	std::vector<ObjectInfo>::iterator i = _waitList.begin();
	while (i != _waitList.end()) {
		Object *obj = i->getObject();
		if (obj && obj == target) {
			find = true;
			_waitResult = *i;
			i = _waitList.erase(i);
		} else {
			i++;
		}
	}
	if (find) {
		_clearWait();
	}
	return find;
}

/**
 * gKɑ΂҂
 * @param name gK
 * @return YIuWFNg҂Ăꍇ true
	 */
bool
Thread::notifyTrigger(const SQChar *name)
{
	bool find = false;
	std::vector<ObjectInfo>::iterator i = _waitList.begin();
	while (i != _waitList.end()) {
		if (i->isSameString(name)) {
			find = true;
			_waitResult = *i;
			i = _waitList.erase(i);
		} else {
			i++;
		}
	}
	if (find) {
		_clearWait();
	}
	return find;
}

// RXgN^
Thread::Thread() : _currentTick(0), _waitTimeout(-1), _status(THREAD_STOP)
{
}

// fXgN^
Thread::~Thread()
{
	_clear();
}

/**
 * j
 */
void
Thread::_clear()
{
	_clearWait();
	_thread.clear();
}

/**
 * IuWFNgɑ΂҂NA
 * @param status LZ̏ꍇ true
 */
void
Thread::_clearWait() {
	std::vector<ObjectInfo>::iterator i = _waitList.begin();
	while (i != _waitList.end()) {
		Object *obj = i->getObject();
		if (obj) {
			obj->removeWait(this);
		}
		i = _waitList.erase(i);
	}
	_waitTimeout = -1;
}

/**
 * ҂o^
 */
SQRESULT
Thread::wait(HSQUIRRELVM v)
{
	_clearWait();
	_waitResult.clear();
	int max = sq_gettop(v);
	for (int i=2;i<=max;i++) {
		switch (sq_gettype(v, 2)) {
		case OT_INTEGER:
		case OT_FLOAT:
			// l̏ꍇ̓^CAEg҂
			{
				int timeout;
				sq_getinteger(v, i, &timeout);
				if (timeout >= 0) {
					if (_waitTimeout < 0  || _waitTimeout > timeout) {
						_waitResult.getStack(v, i);
						_waitTimeout = timeout;
					}
				}
			}
			break;
		case OT_STRING:
			// ҂Xgɓo^
			_waitList.push_back(ObjectInfo(v,i));
			break;
		case OT_INSTANCE:
			// IuWFNgɑ҂o^Ă҂Xgɓo^
			{
				ObjectInfo o(v,i);
				Object *obj = o.getObject();
				if (obj) {
					obj->addWait(this);
				}
				_waitList.push_back(o);
			}
			break;
		}
	}
	return 0;
}

/**
 * wait̃LZ
 */
void
Thread::cancelWait()
{
	_clearWait();
	_waitResult.clear();
}
	
/**
 * sJn
 * @param func sΏۃt@NVB̏ꍇYXNvgǂݍ
 */
SQRESULT
Thread::exec(HSQUIRRELVM v)
{
	_clear();
	
	HSQUIRRELVM testVM = sq_newthread(v, 1024);
	_thread.getStack(v, -1);
	sq_pop(v, 1);
	
	// Xbh擪ɃXNvg[h
	if (sq_gettype(v, 2) == OT_STRING) {
		SQRESULT result;
		if (SQ_FAILED(result = sqstd_loadfile(_thread, getString(v, 2), SQTrue))) {
			_clear();
			return result;
		}
	} else {
		sq_move(_thread, v, 2);
	}
	_status = THREAD_RUN;
	entryThread(v); // X^bN 1ԂɌ̃IuWFNgĂ
	return SQ_OK;
}

/**
 * sI
 */
void
Thread::exit()
{
	_clear();
	notifyAll();
	_status = THREAD_STOP;
}

/**
 * s~
 */
void
Thread::stop()
{
	if (_thread.isThread()) {
		if (_status == THREAD_RUN) {
			_status = THREAD_STOP;
		}
	}
}

/**
 * sĊJ
 */
void
Thread::run()
{
	if (_thread.isThread()) {
		if (_status == THREAD_STOP) {
			_status = THREAD_RUN;
		}
	}
}

/**
 * @return sXe[^X
 */
int
Thread::getStatus()
{
	return !_thread.isThread() ? THREAD_END : isWait() ? THREAD_WAIT : _status;
}

/**
 * Xbh̃C
 * @param diff oߎ
 * @return XbhsIȂ true
 */
bool
Thread::_main(long diff)
{
	_currentTick += diff;
	
	// ^CAEg
	if (_waitTimeout >= 0) {
		_waitTimeout -= diff;
		if (_waitTimeout < 0) {
			_clearWait();
		}
	}
	
	// Xbhs
	if (getStatus() == THREAD_RUN) {
		SQRESULT result;
		if (sq_getvmstate(_thread) == SQ_VMSTATE_SUSPENDED) {
			_waitResult.push(_thread);
			_waitResult.clear();
			result = sq_wakeupvm(_thread, SQTrue, SQFalse, SQTrue);
		} else {
			sq_pushroottable(_thread);
			result = sq_call(_thread, 1, SQFalse, SQTrue);
		}
		if (SQ_FAILED(result)) {
			sq_getlasterror(_thread);
			const SQChar *str;
			if (SQ_SUCCEEDED(sq_getstring(_thread, -1, &str))) {
				PRINT(_thread, str);
			} else {
				PRINT(_thread, _SC("failed to run by unknown reason"));
			}
			sq_pop(_thread, 1);
			exit();
		} else  if (sq_getvmstate(_thread) == SQ_VMSTATE_IDLE) {
			exit();
		}
	}
	return getStatus() == THREAD_END;
}

// -------------------------------------------------------------------------

/*
 * sC[v
 * @param tick tickl
 * ݑ݂Xbh𑍂Ȃ߂łPxsB
 * VXe{̂̃C[v(Cxg{摜)
 * 1xĂяoƂŋ@\Bꂼ̃Xbh́A
 * 疾I suspend() ܂ waitñ\bhĂяoď
 * ̃XbhɈϏKvB
 * @return 
 */
int
Thread::main(long tick)
{
	//dm("tick:" + tick + "\n");
	if (first) {
		prevTick = tick;
		first = false;
	}
	currentTick = tick;
	long diff = tick - prevTick;
	std::vector<ObjectInfo>::iterator i = threadList.begin();
	while (i != threadList.end()) {
		Thread *th = i->getThread();
		if (!th || th->_main(diff)) {
			i = threadList.erase(i);
		} else {
			i++;
		}
	}
	prevTick = tick;
	return threadList.size();
};

/**
 * Xbh̔j
 */
void
Thread::done()
{
	threadList.clear();
}

// -------------------------------------------------------------
//
// O[oXbhp@\
//
// -------------------------------------------------------------

std::vector<ObjectInfo> Thread::threadList; //< Xbhꗗ
bool Thread::first = true;        //< Ăяo
long Thread::prevTick;     //< ǑĂяo
long Thread::currentTick;  //< ̌Ăяo

/**
 * XbhsJno^
 * @param v SquirrelX^bN 1ԂɃXbhIuWFNg͂Ă͂
 */
void
Thread::entryThread(HSQUIRRELVM v)
{
	if (threadList.size() == 0) {
		first = true;
	}
	threadList.push_back(ObjectInfo(v,1));
}

// -------------------------------------------------------------
// O[o\bhp
// -------------------------------------------------------------

/**
 * ݎ̎擾
 */
SQRESULT
Thread::global_getCurrentTick(HSQUIRRELVM v)
{
	sq_pushinteger(v, currentTick);
	return 1;
}

static SQRESULT ERROR_NOTHREAD(HSQUIRRELVM v) {
	return sq_throwerror(v, _SC("no thread"));
}

/*
 * @return ݂̃XbhԂ
 */
SQRESULT
Thread::global_getCurrentThread(HSQUIRRELVM v)
{
	std::vector<ObjectInfo>::iterator i = threadList.begin();
	while (i != threadList.end()) {
		Thread *th = i->getThread();
		if (th && th->isSameThread(v)) {
			i->push(v);
			return 1;
		}
		i++;
	}
	return ERROR_NOTHREAD(v);
}

/*
 * @return ݂̃XbhꗗԂ
 */
SQRESULT
Thread::global_getThreadList(HSQUIRRELVM v)
{
	sq_newarray(v, 0);
	std::vector<ObjectInfo>::const_iterator i = threadList.begin();
	while (i != threadList.end()) {
		i->push(v);
		sq_arrayappend(v, -2);
		i++;
	}
	return 1;
}

/*
 * XNvgVXbhƂĎs
 *  return Thread(func); 
 * @param func XbhŎst@NV
 * @return VXbh
 */
SQRESULT
Thread::global_fork(HSQUIRRELVM v)
{
	//dm("Xbh𐶐!");
	SQRESULT result;
	sq_pushroottable(v); // root
	sq_pushstring(v, SQTHREADNAME, -1);
	if (SQ_SUCCEEDED(result = sq_get(v,-2))) { // class
		sq_pushroottable(v); // :self
		sq_push(v, 2);       // :func
		if (SQ_SUCCEEDED(result = sq_call(v, 2, SQTrue, SQTrue))) { // RXgN^Ăяo
			sq_remove(v, -2); // class, root
			return 1;
		} else {
			sq_pop(v, 1); // class
		}
	}
	sq_pop(v,1); // root
	return result;
}

/**
 * @return ݎs̃XbhIuWFNg(Thread*)
 */
Thread *
Thread::getCurrentThread(HSQUIRRELVM v)
{
	std::vector<ObjectInfo>::iterator i = threadList.begin();
	while (i != threadList.end()) {
		Thread *th = i->getThread();
		if (th && th->isSameThread(v)) {
			return th;
		}
		i++;
	}
	return NULL;
}

/**
 * XNvg؂ւ
 * @param func XbhŎst@NV
 */
SQRESULT
Thread::global_exec(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (th) {
		SQRESULT result;
		if (SQ_FAILED(result = th->exec(v))) {
			return result;
		} else {
			return sq_suspendvm(v);
		}
	} else {
		return ERROR_NOTHREAD(v);
	}
}

/**
 * sXbh̏I
 */
SQRESULT
Thread::global_exit(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (th) {
		th->exit();
		return sq_suspendvm(v);
	} else {
		return ERROR_NOTHREAD(v);
	}
}

/**
 * sXbh̏҂
 * @param target int:ԑ҂(ms), string:gK҂, obj:IuWFNg҂
 * @param timeout ^CAEg(ȗ͖ɑ҂)
 * @return ҂LZꂽ true
 */
SQRESULT
Thread::global_wait(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (th) {
		th->wait(v);
		return sq_suspendvm(v);
	} else {
		return ERROR_NOTHREAD(v);
	}
}

/**
 * SXbhւ̃gKʒm
 * @param name ҂gK
 */
SQRESULT
Thread::global_trigger(HSQUIRRELVM v)
{
	const SQChar *name = getString(v, 2);
	std::vector<ObjectInfo>::iterator i = threadList.begin();
	while (i != threadList.end()) {
		Thread *th = i->getThread();
		if (th) {
			th->notifyTrigger(name);
		}
		i++;
	}
	return SQ_OK;
}

// ------------------------------------------------------------------

/**
 * IuWFNg̃RXgN^
 */
SQRESULT
Thread::constructor(HSQUIRRELVM v)
{
	Thread *self = new Thread();
	if (getParamCount(v) > 1) {
		self->delegate.getStack(v, 3);
	}
	if (getParamCount(v) > 0) {
		self->exec(v);
	}
	return SqPlus::PostConstruct<Thread>(v, self, release);
}

/**
 * NX̓o^
 */
void
Thread::registClass()
{
	// NXo^
	SQCLASS2(Thread,Object);
	SQSVFUNC(Thread,constructor);
	SQVFUNC(Thread,exec);
	SQFUNC(Thread,exit);
	SQFUNC(Thread,stop);
	SQFUNC(Thread,run);
	SQFUNC(Thread,getCurrentTick);
	SQFUNC(Thread,getStatus);
	SQVFUNC(Thread,wait);
	SQFUNC(Thread,cancelWait);
	SQENUM(STOP,THREAD_STOP);
	SQENUM(RUN,THREAD_RUN);
	SQENUM(WAIT,THREAD_WAIT);
	SQENUM(END,THREAD_END);

// \bho^iOj
#define REGISTMETHODNAME(name, method) \
	sq_pushstring(v, _SC(#name), -1);\
	sq_newclosure(v, method, 0);\
	sq_createslot(v, -3);

	// O[o\bh̓o^

	HSQUIRRELVM v = SquirrelVM::GetVMPtr();
	sq_pushroottable(v); // root
	REGISTMETHODNAME(getCurrentTick, global_getCurrentTick);
	REGISTMETHODNAME(getCurrentThread, global_getCurrentThread);
	REGISTMETHODNAME(getThreadList, global_getThreadList);
	REGISTMETHODNAME(fork, global_fork);
	REGISTMETHODNAME(exec, global_exec);
	REGISTMETHODNAME(exit, global_exit);
	REGISTMETHODNAME(wait, global_wait);
	REGISTMETHODNAME(notify, global_trigger);
	sq_pop(v, 1); // root
};

};
